"""Module to facilitate the loading a procesing of data. Largely
inspired/copied from Kipoi (http://github.com/kipoi) and pyTorch
data loaders"""

__all__ = ["extractors", "dataset", "loaders", "samplers", "utils"]